"use strict";
var Actions;
(function (Actions) {
    function getAction(world, config, name) {
        switch (name) {
            case "focus-left": return () => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    const prevColumn = grid.getPrevColumn(column);
                    if (prevColumn === null) {
                        return;
                    }
                    prevColumn.focus();
                });
            };
            case "focus-right": return () => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    const nextColumn = grid.getNextColumn(column);
                    if (nextColumn === null) {
                        return;
                    }
                    nextColumn.focus();
                });
            };
            case "focus-up": return () => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    const prevWindow = column.getPrevWindow(window);
                    if (prevWindow === null) {
                        return;
                    }
                    prevWindow.focus();
                });
            };
            case "focus-down": return () => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    const nextWindow = column.getNextWindow(window);
                    if (nextWindow === null) {
                        return;
                    }
                    nextWindow.focus();
                });
            };
            case "focus-start": return () => {
                world.do((clientManager, desktopManager) => {
                    const grid = desktopManager.getCurrentDesktop().grid;
                    const firstColumn = grid.getFirstColumn();
                    if (firstColumn === null) {
                        return;
                    }
                    firstColumn.focus();
                });
            };
            case "focus-end": return () => {
                world.do((clientManager, desktopManager) => {
                    const grid = desktopManager.getCurrentDesktop().grid;
                    const lastColumn = grid.getLastColumn();
                    if (lastColumn === null) {
                        return;
                    }
                    lastColumn.focus();
                });
            };
            case "window-move-left": return () => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    if (column.getWindowCount() === 1) {
                        // move from own column into existing column
                        const prevColumn = grid.getPrevColumn(column);
                        if (prevColumn === null) {
                            return;
                        }
                        window.moveToColumn(prevColumn);
                        grid.desktop.autoAdjustScroll();
                    }
                    else {
                        // move from shared column into own column
                        const newColumn = new Column(grid, grid.getPrevColumn(column));
                        window.moveToColumn(newColumn);
                    }
                });
            };
            case "window-move-right": return () => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    if (column.getWindowCount() === 1) {
                        // move from own column into existing column
                        const nextColumn = grid.getNextColumn(column);
                        if (nextColumn === null) {
                            return;
                        }
                        window.moveToColumn(nextColumn);
                        grid.desktop.autoAdjustScroll();
                    }
                    else {
                        // move from shared column into own column
                        const newColumn = new Column(grid, column);
                        window.moveToColumn(newColumn);
                    }
                });
            };
            case "window-move-up": return () => {
                // TODO (optimization): only arrange moved windows
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    column.moveWindowUp(window);
                });
            };
            case "window-move-down": return () => {
                // TODO (optimization): only arrange moved windows
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    column.moveWindowDown(window);
                });
            };
            case "window-move-start": return () => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    const newColumn = new Column(grid, null);
                    window.moveToColumn(newColumn);
                });
            };
            case "window-move-end": return () => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    const newColumn = new Column(grid, grid.getLastColumn());
                    window.moveToColumn(newColumn);
                });
            };
            case "window-toggle-floating": return () => {
                const kwinClient = Workspace.activeWindow;
                world.do((clientManager, desktopManager) => {
                    clientManager.toggleFloatingClient(kwinClient);
                });
            };
            case "column-move-left": return () => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    grid.moveColumnLeft(column);
                });
            };
            case "column-move-right": return () => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    grid.moveColumnRight(column);
                });
            };
            case "column-move-start": return () => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    grid.moveColumn(column, null);
                });
            };
            case "column-move-end": return () => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    grid.moveColumn(column, grid.getLastColumn());
                });
            };
            case "column-toggle-stacked": return () => {
                world.doIfTiledFocused(false, (clientManager, desktopManager, window, column, grid) => {
                    column.toggleStacked();
                });
            };
            case "column-width-increase": return () => {
                world.doIfTiledFocused(false, (clientManager, desktopManager, window, column, grid) => {
                    config.columnResizer.increaseWidth(column, config.manualResizeStep);
                });
            };
            case "column-width-decrease": return () => {
                world.doIfTiledFocused(false, (clientManager, desktopManager, window, column, grid) => {
                    config.columnResizer.decreaseWidth(column, config.manualResizeStep);
                });
            };
            case "columns-width-equalize": return () => {
                world.do((clientManager, desktopManager) => {
                    desktopManager.getCurrentDesktop().equalizeVisibleColumnsWidths();
                });
            };
            case "grid-scroll-left": return () => {
                gridScroll(world, -config.manualScrollStep);
            };
            case "grid-scroll-right": return () => {
                gridScroll(world, config.manualScrollStep);
            };
            case "grid-scroll-start": return () => {
                world.do((clientManager, desktopManager) => {
                    const grid = desktopManager.getCurrentDesktop().grid;
                    const firstColumn = grid.getFirstColumn();
                    if (firstColumn === null) {
                        return;
                    }
                    grid.desktop.scrollToColumn(firstColumn);
                });
            };
            case "grid-scroll-end": return () => {
                world.do((clientManager, desktopManager) => {
                    const grid = desktopManager.getCurrentDesktop().grid;
                    const lastColumn = grid.getLastColumn();
                    if (lastColumn === null) {
                        return;
                    }
                    grid.desktop.scrollToColumn(lastColumn);
                });
            };
            case "grid-scroll-focused": return () => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    grid.desktop.scrollCenterRange(column);
                });
            };
            case "grid-scroll-left-column": return () => {
                world.do((clientManager, desktopManager) => {
                    const grid = desktopManager.getCurrentDesktop().grid;
                    const column = grid.getLeftmostVisibleColumn(grid.desktop.getCurrentVisibleRange(), true);
                    if (column === null) {
                        return;
                    }
                    const prevColumn = grid.getPrevColumn(column);
                    if (prevColumn === null) {
                        return;
                    }
                    grid.desktop.scrollToColumn(prevColumn);
                });
            };
            case "grid-scroll-right-column": return () => {
                world.do((clientManager, desktopManager) => {
                    const grid = desktopManager.getCurrentDesktop().grid;
                    const column = grid.getRightmostVisibleColumn(grid.desktop.getCurrentVisibleRange(), true);
                    if (column === null) {
                        return;
                    }
                    const nextColumn = grid.getNextColumn(column);
                    if (nextColumn === null) {
                        return;
                    }
                    grid.desktop.scrollToColumn(nextColumn);
                });
            };
            case "screen-switch": return () => {
                world.do((clientManager, desktopManager) => {
                    desktopManager.selectScreen(Workspace.activeScreen);
                });
            };
            default: throw new Error("unknown action: " + name);
        }
    }
    Actions.getAction = getAction;
    function getNumAction(world, name) {
        switch (name) {
            case "focus-": return (columnIndex) => {
                world.do((clientManager, desktopManager) => {
                    const grid = desktopManager.getCurrentDesktop().grid;
                    const targetColumn = grid.getColumnAtIndex(columnIndex);
                    if (targetColumn === null) {
                        return;
                    }
                    targetColumn.focus();
                });
            };
            case "window-move-to-column-": return (columnIndex) => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    const targetColumn = grid.getColumnAtIndex(columnIndex);
                    if (targetColumn === null) {
                        return;
                    }
                    window.moveToColumn(targetColumn);
                    grid.desktop.autoAdjustScroll();
                });
            };
            case "column-move-to-column-": return (columnIndex) => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, grid) => {
                    const targetColumn = grid.getColumnAtIndex(columnIndex);
                    if (targetColumn === null || targetColumn === column) {
                        return;
                    }
                    if (targetColumn.isToTheRightOf(column)) {
                        grid.moveColumn(column, targetColumn);
                    }
                    else {
                        grid.moveColumn(column, grid.getPrevColumn(targetColumn));
                    }
                });
            };
            case "column-move-to-desktop-": return (desktopIndex) => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, oldGrid) => {
                    const kwinDesktop = Workspace.desktops[desktopIndex];
                    if (kwinDesktop === undefined) {
                        return;
                    }
                    const newGrid = desktopManager.getDesktopInCurrentActivity(kwinDesktop).grid;
                    if (newGrid === null || newGrid === oldGrid) {
                        return;
                    }
                    column.moveToGrid(newGrid, newGrid.getLastColumn());
                });
            };
            case "tail-move-to-desktop-": return (desktopIndex) => {
                world.doIfTiledFocused(true, (clientManager, desktopManager, window, column, oldGrid) => {
                    const kwinDesktop = Workspace.desktops[desktopIndex];
                    if (kwinDesktop === undefined) {
                        return;
                    }
                    const newGrid = desktopManager.getDesktopInCurrentActivity(kwinDesktop).grid;
                    if (newGrid === null || newGrid === oldGrid) {
                        return;
                    }
                    oldGrid.evacuateTail(newGrid, column);
                });
            };
            default: throw new Error("unknown num action: " + name);
        }
    }
    Actions.getNumAction = getNumAction;
    function gridScroll(world, amount) {
        world.do((clientManager, desktopManager) => {
            const grid = desktopManager.getCurrentDesktop().grid;
            grid.desktop.adjustScroll(amount, false);
        });
    }
})(Actions || (Actions = {}));
function initWorkspaceSignalHandlers(world) {
    const manager = new SignalManager();
    manager.connect(Workspace.windowAdded, (kwinClient) => {
        if (Clients.canTileEver(kwinClient)) {
            // never open new tileable clients on all desktops or activities
            Clients.makeTileable(kwinClient);
        }
        world.do((clientManager, desktopManager) => {
            clientManager.addClient(kwinClient);
        });
    });
    manager.connect(Workspace.windowRemoved, (kwinClient) => {
        world.do((clientManager, desktopManager) => {
            clientManager.removeClient(kwinClient, true);
        });
    });
    manager.connect(Workspace.windowActivated, (kwinClient) => {
        if (kwinClient === null) {
            return;
        }
        world.do((clientManager, desktopManager) => {
            clientManager.onClientFocused(kwinClient);
        });
    });
    manager.connect(Workspace.currentDesktopChanged, () => {
        world.do(() => { }); // re-arrange desktop
    });
    manager.connect(Workspace.currentActivityChanged, () => {
        world.do(() => { }); // re-arrange desktop
    });
    manager.connect(Workspace.screensChanged, () => {
        world.do((clientManager, desktopManager) => {
            desktopManager.selectScreen(Workspace.activeScreen);
        });
    });
    manager.connect(Workspace.activitiesChanged, () => {
        world.do((clientManager, desktopManager) => {
            desktopManager.updateActivities();
        });
    });
    manager.connect(Workspace.desktopsChanged, () => {
        world.do((clientManager, desktopManager) => {
            desktopManager.updateDesktops();
        });
    });
    manager.connect(Workspace.virtualScreenSizeChanged, () => {
        world.onScreenResized();
    });
    return manager;
}
class ContextualResizer {
    increaseWidth(column, step) {
        const grid = column.grid;
        const desktop = grid.desktop;
        const visibleRange = desktop.getCurrentVisibleRange();
        if (!column.isVisible(visibleRange, true) || column.getWidth() >= column.getMaxWidth()) {
            return;
        }
        let leftVisibleColumn = grid.getLeftmostVisibleColumn(visibleRange, true);
        let rightVisibleColumn = grid.getRightmostVisibleColumn(visibleRange, true);
        if (leftVisibleColumn === null || rightVisibleColumn === null) {
            console.assert(false); // should at least see self
            return;
        }
        const leftSpace = leftVisibleColumn.getLeft() - visibleRange.getLeft();
        const rightSpace = visibleRange.getRight() - rightVisibleColumn.getRight();
        const newWidth = ContextualResizer.findNextStep([
            visibleRange.getWidth(),
            column.getWidth() + step,
            column.getWidth() + leftSpace + rightSpace,
            column.getWidth() + leftSpace + rightSpace + leftVisibleColumn.getWidth() + grid.config.gapsInnerHorizontal,
            column.getWidth() + leftSpace + rightSpace + rightVisibleColumn.getWidth() + grid.config.gapsInnerHorizontal,
        ], width => width - column.getWidth());
        if (newWidth === undefined) {
            return;
        }
        column.setWidth(newWidth, true);
        desktop.scrollCenterVisible(column);
    }
    decreaseWidth(column, step) {
        const grid = column.grid;
        const desktop = grid.desktop;
        const visibleRange = desktop.getCurrentVisibleRange();
        if (!column.isVisible(visibleRange, true) || column.getWidth() <= column.getMinWidth()) {
            return;
        }
        const leftVisibleColumn = grid.getLeftmostVisibleColumn(visibleRange, true);
        const rightVisibleColumn = grid.getRightmostVisibleColumn(visibleRange, true);
        if (leftVisibleColumn === null || rightVisibleColumn === null) {
            console.assert(false); // should at least see self
            return;
        }
        let leftOffScreenColumn = grid.getPrevColumn(leftVisibleColumn);
        if (leftOffScreenColumn === column) {
            leftOffScreenColumn = null;
        }
        let rightOffScreenColumn = grid.getNextColumn(rightVisibleColumn);
        if (rightOffScreenColumn === column) {
            rightOffScreenColumn = null;
        }
        const visibleColumnsWidth = rightVisibleColumn.getRight() - leftVisibleColumn.getLeft();
        const unusedWidth = visibleRange.getWidth() - visibleColumnsWidth;
        const leftOffScreen = leftOffScreenColumn === null ? 0 : leftOffScreenColumn.getWidth() + grid.config.gapsInnerHorizontal - unusedWidth;
        const rightOffScreen = rightOffScreenColumn === null ? 0 : rightOffScreenColumn.getWidth() + grid.config.gapsInnerHorizontal - unusedWidth;
        const newWidth = ContextualResizer.findNextStep([
            visibleRange.getWidth(),
            column.getWidth() - step,
            column.getWidth() - leftOffScreen,
            column.getWidth() - rightOffScreen,
        ], width => column.getWidth() - width);
        if (newWidth === undefined) {
            return;
        }
        column.setWidth(newWidth, true);
        desktop.scrollCenterVisible(column);
    }
    static findNextStep(steps, evaluate) {
        let bestScore = Infinity;
        let bestStep = undefined;
        for (const step of steps) {
            const score = evaluate(step);
            if (score > 0 && score < bestScore) {
                bestScore = score;
                bestStep = step;
            }
        }
        return bestStep;
    }
}
class RawResizer {
    increaseWidth(column, step) {
        column.adjustWidth(step, true);
    }
    decreaseWidth(column, step) {
        column.adjustWidth(-step, true);
    }
}
class CenterClamper {
    clampScrollX(desktop, x) {
        const firstColumn = desktop.grid.getFirstColumn();
        if (firstColumn === null) {
            return 0;
        }
        const lastColumn = desktop.grid.getLastColumn();
        let minScroll = Math.round((firstColumn.getWidth() - desktop.tilingArea.width) / 2);
        let maxScroll = Math.round(desktop.grid.getWidth() - (desktop.tilingArea.width + lastColumn.getWidth()) / 2);
        return clamp(x, minScroll, maxScroll);
    }
}
class EdgeClamper {
    clampScrollX(desktop, x) {
        let minScroll = 0;
        let maxScroll = desktop.grid.getWidth() - desktop.tilingArea.width;
        if (maxScroll < 0) {
            return Math.round(maxScroll / 2);
        }
        return clamp(x, minScroll, maxScroll);
    }
}
class CenteredScroller {
    scrollToColumn(desktop, column) {
        desktop.scrollCenterRange(column);
    }
}
class GroupedScroller {
    scrollToColumn(desktop, column) {
        desktop.scrollCenterVisible(column);
    }
}
class LazyScroller {
    scrollToColumn(desktop, column) {
        desktop.scrollIntoView(column);
    }
}
const defaultWindowRules = `[
    {
        "class": "ksmserver-logout-greeter",
        "tile": false
    },
    {
        "class": "xwaylandvideobridge",
        "tile": false
    },
    {
        "class": "(org\\\\.kde\\\\.)?plasmashell",
        "tile": false
    },
    {
        "class": "(org\\\\.kde\\\\.)?kded6",
        "tile": false
    },
    {
        "class": "(org\\\\.kde\\\\.)?kcalc",
        "tile": false
    },
    {
        "class": "(org\\\\.kde\\\\.)?kfind",
        "tile": true
    },
    {
        "class": "(org\\\\.kde\\\\.)?kruler",
        "tile": false
    },
    {
        "class": "(org\\\\.kde\\\\.)?krunner",
        "tile": false
    },
    {
        "class": "(org\\\\.kde\\\\.)?yakuake",
        "tile": false
    },
    {
        "class": "zoom",
        "caption": "Zoom Cloud Meetings|zoom|zoom <2>",
        "tile": false
    },
    {
        "class": "jetbrains-idea",
        "caption": "splash",
        "tile": false
    },
    {
        "class": "jetbrains-studio",
        "caption": "splash",
        "tile": false
    },
    {
        "class": "jetbrains-idea",
        "caption": "Unstash Changes|Paths Affected by stash@.*",
        "tile": true
    },
    {
        "class": "jetbrains-studio",
        "caption": "Unstash Changes|Paths Affected by stash@.*",
        "tile": true
    }
]`;
const configDef = [
    {
        name: "gapsOuterTop",
        type: "UInt",
        default: 16,
    },
    {
        name: "gapsOuterBottom",
        type: "UInt",
        default: 16,
    },
    {
        name: "gapsOuterLeft",
        type: "UInt",
        default: 16,
    },
    {
        name: "gapsOuterRight",
        type: "UInt",
        default: 16,
    },
    {
        name: "gapsInnerHorizontal",
        type: "UInt",
        default: 8,
    },
    {
        name: "gapsInnerVertical",
        type: "UInt",
        default: 8,
    },
    {
        name: "manualScrollStep",
        type: "UInt",
        default: 200,
    },
    {
        name: "manualResizeStep",
        type: "UInt",
        default: 600,
    },
    {
        name: "offScreenOpacity",
        type: "UInt",
        default: 100,
    },
    {
        name: "untileOnDrag",
        type: "Bool",
        default: true,
    },
    {
        name: "stackColumnsByDefault",
        type: "Bool",
        default: false,
    },
    {
        name: "resizeNeighborColumn",
        type: "Bool",
        default: false,
    },
    {
        name: "reMaximize",
        type: "Bool",
        default: false,
    },
    {
        name: "skipSwitcher",
        type: "Bool",
        default: false,
    },
    {
        name: "scrollingLazy",
        type: "Bool",
        default: true,
    },
    {
        name: "scrollingCentered",
        type: "Bool",
        default: false,
    },
    {
        name: "scrollingGrouped",
        type: "Bool",
        default: false,
    },
    {
        name: "tiledKeepBelow",
        type: "Bool",
        default: true,
    },
    {
        name: "floatingKeepAbove",
        type: "Bool",
        default: false,
    },
    {
        name: "noLayering",
        type: "Bool",
        default: false,
    },
    {
        name: "windowRules",
        type: "String",
        default: defaultWindowRules,
    }
];
function loadConfig() {
    const config = {};
    for (const entry of configDef) {
        config[entry.name] = KWin.readConfig(entry.name, entry.default);
    }
    return config;
}
const keyBindings = [
    {
        name: "window-toggle-floating",
        description: "Toggle floating",
        defaultKeySequence: "Meta+Space",
    },
    {
        name: "focus-left",
        description: "Move focus left",
        defaultKeySequence: "Meta+A",
    },
    {
        name: "focus-right",
        description: "Move focus right",
        comment: "Clashes with default KDE shortcuts, may require manual remapping",
        defaultKeySequence: "Meta+D",
    },
    {
        name: "focus-up",
        description: "Move focus up",
        comment: "Clashes with default KDE shortcuts, may require manual remapping",
        defaultKeySequence: "Meta+W",
    },
    {
        name: "focus-down",
        description: "Move focus down",
        comment: "Clashes with default KDE shortcuts, may require manual remapping",
        defaultKeySequence: "Meta+S",
    },
    {
        name: "focus-start",
        description: "Move focus to start",
        defaultKeySequence: "Meta+Home",
    },
    {
        name: "focus-end",
        description: "Move focus to end",
        defaultKeySequence: "Meta+End",
    },
    {
        name: "window-move-left",
        description: "Move window left",
        comment: "Moves window out of and into columns",
        defaultKeySequence: "Meta+Shift+A",
    },
    {
        name: "window-move-right",
        description: "Move window right",
        comment: "Moves window out of and into columns",
        defaultKeySequence: "Meta+Shift+D",
    },
    {
        name: "window-move-up",
        description: "Move window up",
        defaultKeySequence: "Meta+Shift+W",
    },
    {
        name: "window-move-down",
        description: "Move window down",
        defaultKeySequence: "Meta+Shift+S",
    },
    {
        name: "window-move-start",
        description: "Move window to start",
        defaultKeySequence: "Meta+Shift+Home",
    },
    {
        name: "window-move-end",
        description: "Move window to end",
        defaultKeySequence: "Meta+Shift+End",
    },
    {
        name: "column-toggle-stacked",
        description: "Toggle stacked layout for focused column",
        comment: "One window in the column visible, others shaded; not supported on Wayland",
        defaultKeySequence: "Meta+X",
    },
    {
        name: "column-move-left",
        description: "Move column left",
        defaultKeySequence: "Meta+Ctrl+Shift+A",
    },
    {
        name: "column-move-right",
        description: "Move column right",
        defaultKeySequence: "Meta+Ctrl+Shift+D",
    },
    {
        name: "column-move-start",
        description: "Move column to start",
        defaultKeySequence: "Meta+Ctrl+Shift+Home",
    },
    {
        name: "column-move-end",
        description: "Move column to end",
        defaultKeySequence: "Meta+Ctrl+Shift+End",
    },
    {
        name: "column-width-increase",
        description: "Increase column width",
        defaultKeySequence: "Meta+Ctrl++",
    },
    {
        name: "column-width-decrease",
        description: "Decrease column width",
        defaultKeySequence: "Meta+Ctrl+-",
    },
    {
        name: "columns-width-equalize",
        description: "Equalize widths of visible columns",
        defaultKeySequence: "Meta+Ctrl+X",
    },
    {
        name: "grid-scroll-focused",
        description: "Center focused window",
        comment: "Scrolls so that the focused window is centered in the screen",
        defaultKeySequence: "Meta+Alt+Return",
    },
    {
        name: "grid-scroll-left-column",
        description: "Scroll one column to the left",
        defaultKeySequence: "Meta+Alt+A",
    },
    {
        name: "grid-scroll-right-column",
        description: "Scroll one column to the right",
        defaultKeySequence: "Meta+Alt+D",
    },
    {
        name: "grid-scroll-left",
        description: "Scroll left",
        defaultKeySequence: "Meta+Alt+PgUp",
    },
    {
        name: "grid-scroll-right",
        description: "Scroll right",
        defaultKeySequence: "Meta+Alt+PgDown",
    },
    {
        name: "grid-scroll-start",
        description: "Scroll to start",
        defaultKeySequence: "Meta+Alt+Home",
    },
    {
        name: "grid-scroll-end",
        description: "Scroll to end",
        defaultKeySequence: "Meta+Alt+End",
    },
    {
        name: "screen-switch",
        description: "Move Karousel grid to the current screen",
        defaultKeySequence: "Meta+Ctrl+Return",
    }
];
const numKeyBindings = [
    {
        name: "focus-",
        description: "Move focus to column ",
        comment: "Clashes with default KDE shortcuts, may require manual remapping",
        defaultModifiers: "Meta",
        fKeys: false,
    },
    {
        name: "window-move-to-column-",
        description: "Move window to column ",
        comment: "Requires manual remapping according to your keyboard layout, e.g. Meta+Shift+1 -> Meta+!",
        defaultModifiers: "Meta+Shift",
        fKeys: false,
    },
    {
        name: "column-move-to-column-",
        description: "Move column to position ",
        comment: "Requires manual remapping according to your keyboard layout, e.g. Meta+Ctrl+Shift+1 -> Meta+Ctrl+!",
        defaultModifiers: "Meta+Ctrl+Shift",
        fKeys: false,
    },
    {
        name: "column-move-to-desktop-",
        description: "Move column to desktop ",
        defaultModifiers: "Meta+Ctrl+Shift",
        fKeys: true,
    },
    {
        name: "tail-move-to-desktop-",
        description: "Move this and all following columns to desktop ",
        defaultModifiers: "Meta+Ctrl+Shift+Alt",
        fKeys: true,
    },
];
function catchWrap(f) {
    return () => {
        try {
            f();
        }
        catch (error) {
            log(error);
            log(error.stack);
        }
    };
}
function registerKeyBinding(world, config, shortcutActions, keyBinding) {
    shortcutActions.push(new ShortcutAction(keyBinding, catchWrap(Actions.getAction(world, config, keyBinding.name))));
}
function registerNumKeyBindings(world, shortcutActions, numKeyBinding) {
    const numPrefix = numKeyBinding.fKeys ? "F" : "";
    const n = numKeyBinding.fKeys ? 12 : 9;
    for (let i = 0; i < 12; i++) {
        const numKey = String(i + 1);
        const keySequence = i < n ?
            numKeyBinding.defaultModifiers + "+" + numPrefix + numKey :
            "";
        const action = Actions.getNumAction(world, numKeyBinding.name);
        shortcutActions.push(new ShortcutAction({
            name: numKeyBinding.name + numKey,
            description: numKeyBinding.description + numKey,
            defaultKeySequence: keySequence,
        }, catchWrap(() => action(i))));
    }
}
// TODO: refactor
function registerKeyBindings(world, config) {
    const shortcutActions = [];
    for (const keyBinding of keyBindings) {
        registerKeyBinding(world, config, shortcutActions, keyBinding);
    }
    for (const numKeyBinding of numKeyBindings) {
        registerNumKeyBindings(world, shortcutActions, numKeyBinding);
    }
    return shortcutActions;
}
class Column {
    constructor(grid, prevColumn) {
        this.gridX = 0;
        this.width = 0;
        this.windows = new LinkedList();
        this.stacked = grid.config.stackColumnsByDefault;
        this.focusTaker = null;
        this.grid = grid;
        this.grid.onColumnAdded(this, prevColumn);
    }
    moveToGrid(targetGrid, prevColumn) {
        if (targetGrid === this.grid) {
            this.grid.moveColumn(this, prevColumn);
        }
        else {
            this.grid.onColumnRemoved(this, false);
            this.grid = targetGrid;
            targetGrid.onColumnAdded(this, prevColumn);
            for (const window of this.windows.iterator()) {
                window.client.kwinClient.desktops = [targetGrid.desktop.kwinDesktop];
            }
        }
    }
    isToTheLeftOf(other) {
        return this.gridX < other.gridX;
    }
    isToTheRightOf(other) {
        return this.gridX > other.gridX;
    }
    moveWindowUp(window) {
        this.windows.moveBack(window);
        this.grid.desktop.onLayoutChanged();
    }
    moveWindowDown(window) {
        this.windows.moveForward(window);
        this.grid.desktop.onLayoutChanged();
    }
    getWindowCount() {
        return this.windows.length();
    }
    isEmpty() {
        return this.getWindowCount() === 0;
    }
    getPrevWindow(window) {
        return this.windows.getPrev(window);
    }
    getNextWindow(window) {
        return this.windows.getNext(window);
    }
    getWidth() {
        return this.width;
    }
    getMinWidth() {
        let maxMinWidth = Column.minWidth;
        for (const window of this.windows.iterator()) {
            const minWidth = window.client.kwinClient.minSize.width;
            if (minWidth > maxMinWidth) {
                maxMinWidth = minWidth;
            }
        }
        return maxMinWidth;
    }
    getMaxWidth() {
        return this.grid.desktop.tilingArea.width;
    }
    setWidth(width, setPreferred) {
        width = clamp(width, this.getMinWidth(), this.getMaxWidth());
        if (width === this.width) {
            return;
        }
        this.width = width;
        if (setPreferred) {
            for (const window of this.windows.iterator()) {
                window.client.preferredWidth = width;
            }
        }
        this.grid.onColumnWidthChanged(this);
    }
    adjustWidth(widthDelta, setPreferred) {
        this.setWidth(this.width + widthDelta, setPreferred);
    }
    updateWidth() {
        let minErr = Infinity;
        let closestPreferredWidth = this.width;
        for (const window of this.windows.iterator()) {
            const err = Math.abs(window.client.preferredWidth - this.width);
            if (err < minErr) {
                minErr = err;
                closestPreferredWidth = window.client.preferredWidth;
            }
        }
        this.setWidth(closestPreferredWidth, false);
    }
    // returns x position of left edge in grid space
    getLeft() {
        return this.gridX;
    }
    // returns x position of right edge in grid space
    getRight() {
        return this.gridX + this.width;
    }
    adjustWindowHeight(window, heightDelta, top) {
        const otherWindow = top ? this.windows.getPrev(window) : this.windows.getNext(window);
        if (otherWindow === null) {
            return;
        }
        window.height += heightDelta;
        otherWindow.height -= heightDelta;
        this.grid.desktop.onLayoutChanged();
    }
    resizeWindows() {
        const nWindows = this.windows.length();
        if (nWindows === 0) {
            return;
        }
        if (nWindows === 1) {
            this.stacked = this.grid.config.stackColumnsByDefault;
        }
        let remainingPixels = this.grid.desktop.tilingArea.height - (nWindows - 1) * this.grid.config.gapsInnerVertical;
        let remainingWindows = nWindows;
        for (const window of this.windows.iterator()) {
            const windowHeight = Math.round(remainingPixels / remainingWindows);
            window.height = windowHeight;
            remainingPixels -= windowHeight;
            remainingWindows--;
        }
        // TODO: respect min height
        this.grid.desktop.onLayoutChanged();
    }
    getFocusTaker() {
        if (this.focusTaker === null || !this.windows.contains(this.focusTaker)) {
            return null;
        }
        return this.focusTaker;
    }
    focus() {
        const window = this.getFocusTaker() ?? this.windows.getFirst();
        if (window === null) {
            return;
        }
        window.focus();
    }
    arrange(x, visibleRange, forceOpaque) {
        if (this.grid.config.offScreenOpacity < 1.0 && !forceOpaque) {
            const opacity = this.isVisible(visibleRange, true) ? 100 : this.grid.config.offScreenOpacity;
            for (const window of this.windows.iterator()) {
                window.client.kwinClient.opacity = opacity;
            }
        }
        if (this.stacked && this.windows.length() >= 2 && this.canStack()) {
            this.arrangeStacked(x);
            return;
        }
        let y = this.grid.desktop.tilingArea.y;
        for (const window of this.windows.iterator()) {
            window.client.setShade(false);
            window.arrange(x, y, this.width, window.height);
            y += window.height + this.grid.config.gapsInnerVertical;
        }
    }
    arrangeStacked(x) {
        const expandedWindow = this.getFocusTaker();
        let collapsedHeight;
        for (const window of this.windows.iterator()) {
            if (window === expandedWindow) {
                window.client.setShade(false);
            }
            else {
                window.client.setShade(true);
                collapsedHeight = window.client.kwinClient.frameGeometry.height;
            }
        }
        const nCollapsed = this.getWindowCount() - 1;
        const expandedHeight = this.grid.desktop.tilingArea.height - nCollapsed * (collapsedHeight + this.grid.config.gapsInnerVertical);
        let y = this.grid.desktop.tilingArea.y;
        for (const window of this.windows.iterator()) {
            if (window === expandedWindow) {
                window.arrange(x, y, this.width, expandedHeight);
                y += expandedHeight;
            }
            else {
                window.arrange(x, y, this.width, window.height);
                y += collapsedHeight;
            }
            y += this.grid.config.gapsInnerVertical;
        }
    }
    toggleStacked() {
        if (this.windows.length() < 2) {
            return;
        }
        this.stacked = !this.stacked;
        this.grid.desktop.onLayoutChanged();
    }
    canStack() {
        for (const window of this.windows.iterator()) {
            if (!window.client.kwinClient.shadeable) {
                return false;
            }
        }
        return true;
    }
    isVisible(visibleRange, fullyVisible) {
        if (fullyVisible) {
            return this.getLeft() >= visibleRange.getLeft() &&
                this.getRight() <= visibleRange.getRight();
        }
        else {
            return this.getRight() + this.grid.config.gapsInnerHorizontal > visibleRange.getLeft() &&
                this.getLeft() - this.grid.config.gapsInnerHorizontal < visibleRange.getRight();
        }
    }
    onWindowAdded(window) {
        this.windows.insertEnd(window);
        if (this.width === 0) {
            this.setWidth(window.client.preferredWidth, false);
        }
        // TODO: also change column width if the new window requires it
        this.resizeWindows();
        if (window.isFocused()) {
            this.onWindowFocused(window);
        }
        this.grid.desktop.onLayoutChanged();
    }
    onWindowRemoved(window, passFocus) {
        const lastWindow = this.windows.length() === 1;
        const windowToFocus = this.getPrevWindow(window) ?? this.getNextWindow(window);
        this.windows.remove(window);
        if (window === this.focusTaker) {
            this.focusTaker = windowToFocus;
        }
        if (lastWindow) {
            console.assert(this.isEmpty());
            this.destroy(passFocus);
        }
        else {
            this.resizeWindows();
            if (passFocus && windowToFocus !== null) {
                windowToFocus.focus();
            }
        }
        this.grid.desktop.onLayoutChanged();
    }
    onWindowFocused(window) {
        this.grid.onColumnFocused(this);
        this.focusTaker = window;
    }
    restoreToTiled() {
        const lastFocusedWindow = this.getFocusTaker();
        if (lastFocusedWindow !== null) {
            lastFocusedWindow.restoreToTiled();
        }
    }
    destroy(passFocus) {
        this.grid.onColumnRemoved(this, passFocus);
    }
}
Column.minWidth = 10;
class Desktop {
    constructor(kwinDesktop, pinManager, config, getScreen, layoutConfig) {
        this.kwinDesktop = kwinDesktop;
        this.pinManager = pinManager;
        this.config = config;
        this.getScreen = getScreen;
        this.scrollX = 0;
        this.dirty = true;
        this.dirtyScroll = true;
        this.dirtyPins = true;
        this.grid = new Grid(this, layoutConfig);
        this.clientArea = Desktop.getClientArea(this.getScreen(), kwinDesktop);
        this.tilingArea = Desktop.getTilingArea(this.clientArea, kwinDesktop, pinManager, config);
    }
    updateArea() {
        const newClientArea = Desktop.getClientArea(this.getScreen(), this.kwinDesktop);
        if (newClientArea === this.clientArea && !this.dirtyPins) {
            return;
        }
        this.clientArea = newClientArea;
        this.tilingArea = Desktop.getTilingArea(newClientArea, this.kwinDesktop, this.pinManager, this.config);
        this.dirty = true;
        this.dirtyScroll = true;
        this.dirtyPins = false;
        this.grid.onScreenSizeChanged();
        this.autoAdjustScroll();
    }
    static getClientArea(screen, kwinDesktop) {
        return Workspace.clientArea(0 /* ClientAreaOption.PlacementArea */, screen, kwinDesktop);
    }
    static getTilingArea(clientArea, kwinDesktop, pinManager, config) {
        const availableSpace = pinManager.getAvailableSpace(kwinDesktop, clientArea);
        const top = availableSpace.top + config.marginTop;
        const bottom = availableSpace.bottom - config.marginBottom;
        const left = availableSpace.left + config.marginLeft;
        const right = availableSpace.right - config.marginRight;
        return Qt.rect(left, top, right - left, bottom - top);
    }
    scrollIntoView(range) {
        const left = range.getLeft();
        const right = range.getRight();
        const initialVisibleRange = this.getCurrentVisibleRange();
        let targetScrollX;
        if (left < initialVisibleRange.getLeft()) {
            targetScrollX = left;
        }
        else if (right > initialVisibleRange.getRight()) {
            targetScrollX = right - this.tilingArea.width;
        }
        else {
            targetScrollX = initialVisibleRange.getLeft();
        }
        this.setScroll(targetScrollX, false);
    }
    scrollCenterRange(range) {
        const windowCenter = range.getLeft() + range.getWidth() / 2;
        const screenCenter = this.scrollX + this.tilingArea.width / 2;
        this.adjustScroll(Math.round(windowCenter - screenCenter), false);
    }
    scrollCenterVisible(focusedColumn) {
        const columnRange = new Desktop.ColumnRange(focusedColumn);
        const visibleRange = this.getCurrentVisibleRange();
        columnRange.addNeighbors(visibleRange, this.grid.config.gapsInnerHorizontal);
        this.scrollCenterRange(columnRange);
    }
    autoAdjustScroll() {
        const focusedColumn = this.grid.getLastFocusedColumn();
        if (focusedColumn === null || focusedColumn.grid !== this.grid) {
            return;
        }
        this.scrollToColumn(focusedColumn);
    }
    scrollToColumn(column) {
        if (this.dirtyScroll || !column.isVisible(this.getCurrentVisibleRange(), true)) {
            this.config.scroller.scrollToColumn(this, column);
        }
    }
    getVisibleRange(scrollX) {
        return new Desktop.RangeImpl(scrollX, this.tilingArea.width);
    }
    getCurrentVisibleRange() {
        return this.getVisibleRange(this.scrollX);
    }
    clampScrollX(x) {
        return this.config.clamper.clampScrollX(this, x);
    }
    setScroll(x, force) {
        const oldScrollX = this.scrollX;
        this.scrollX = force ? x : this.clampScrollX(x);
        if (this.scrollX !== oldScrollX) {
            this.onLayoutChanged();
        }
        this.dirtyScroll = false;
    }
    adjustScroll(dx, force) {
        this.setScroll(this.scrollX + dx, force);
    }
    equalizeVisibleColumnsWidths() {
        const visibleRange = this.getCurrentVisibleRange();
        const visibleColumns = Array.from(this.grid.getVisibleColumns(visibleRange, true));
        let remainingWidth = this.tilingArea.width - (visibleColumns.length - 1) * this.grid.config.gapsInnerHorizontal;
        let remainingColumns = visibleColumns.length;
        const minWidths = visibleColumns.map(column => column.getMinWidth()).sort((a, b) => b - a);
        for (const minWidth of minWidths) {
            if (minWidth > remainingWidth / remainingColumns) {
                remainingWidth -= minWidth;
                remainingColumns--;
            }
        }
        const avgWidth = remainingWidth / remainingColumns;
        for (const column of visibleColumns) {
            const minWidth = column.getMinWidth();
            if (minWidth > avgWidth) {
                column.setWidth(minWidth, true);
            }
            else {
                const columnWidth = Math.round(remainingWidth / remainingColumns);
                column.setWidth(columnWidth, true);
                remainingWidth -= column.getWidth();
                remainingColumns--;
            }
        }
        this.scrollCenterRange(Desktop.RangeImpl.fromRanges(visibleColumns[0], visibleColumns[visibleColumns.length - 1]));
    }
    arrange() {
        // TODO (optimization): only arrange visible windows
        this.updateArea();
        if (!this.dirty) {
            return;
        }
        this.grid.arrange(this.tilingArea.x - this.scrollX, this.getCurrentVisibleRange());
        this.dirty = false;
    }
    onLayoutChanged() {
        this.dirty = true;
        this.dirtyScroll = true;
    }
    onPinsChanged() {
        this.dirty = true;
        this.dirtyScroll = true;
        this.dirtyPins = true;
    }
    destroy() {
        this.grid.destroy();
    }
}
(function (Desktop) {
    class RangeImpl {
        constructor(x, width) {
            this.x = x;
            this.width = width;
        }
        getLeft() {
            return this.x;
        }
        getRight() {
            return this.x + this.width;
        }
        getWidth() {
            return this.width;
        }
        static fromRanges(leftRange, rightRange) {
            const left = leftRange.getLeft();
            const right = rightRange.getRight();
            return new RangeImpl(left, right - left);
        }
    }
    Desktop.RangeImpl = RangeImpl;
    class ColumnRange {
        constructor(initialColumn) {
            this.left = initialColumn;
            this.right = initialColumn;
            this.width = initialColumn.getWidth();
        }
        addNeighbors(visibleRange, gap) {
            const grid = this.left.grid;
            const columnRange = this;
            function canFit(column) {
                return columnRange.width + gap + column.getWidth() <= visibleRange.getWidth();
            }
            function isUsable(column) {
                return column !== null && canFit(column);
            }
            let leftColumn = grid.getPrevColumn(this.left);
            let rightColumn = grid.getNextColumn(this.right);
            function checkColumns() {
                if (!isUsable(leftColumn)) {
                    leftColumn = null;
                }
                if (!isUsable(rightColumn)) {
                    rightColumn = null;
                }
            }
            checkColumns();
            const visibleCenter = visibleRange.getLeft() + visibleRange.getWidth() / 2;
            while (leftColumn !== null || rightColumn !== null) {
                const leftToCenter = leftColumn === null ? Infinity : Math.abs(leftColumn.getLeft() - visibleCenter);
                const rightToCenter = rightColumn === null ? Infinity : Math.abs(rightColumn.getRight() - visibleCenter);
                if (leftToCenter < rightToCenter) {
                    this.addLeft(leftColumn, gap);
                    leftColumn = grid.getPrevColumn(leftColumn);
                }
                else {
                    this.addRight(rightColumn, gap);
                    rightColumn = grid.getNextColumn(rightColumn);
                }
                checkColumns();
            }
        }
        addLeft(column, gap) {
            this.left = column;
            this.width += column.getWidth() + gap;
        }
        addRight(column, gap) {
            this.right = column;
            this.width += column.getWidth() + gap;
        }
        getLeft() {
            return this.left.getLeft();
        }
        getRight() {
            return this.right.getRight();
        }
        getWidth() {
            return this.width;
        }
    }
    Desktop.ColumnRange = ColumnRange;
})(Desktop || (Desktop = {}));
class Grid {
    constructor(desktop, config) {
        this.desktop = desktop;
        this.config = config;
        this.columns = new LinkedList();
        this.lastFocusedColumn = null;
        this.width = 0;
        this.userResize = false;
        this.userResizeFinishedDelayer = new Delayer(50, () => {
            // this delay prevents windows' contents from freezing after resizing
            this.desktop.onLayoutChanged();
            this.desktop.autoAdjustScroll();
            this.desktop.arrange();
        });
    }
    moveColumn(column, prevColumn) {
        if (column === prevColumn) {
            return;
        }
        const movedLeft = prevColumn === null ? true : column.isToTheRightOf(prevColumn);
        const firstMovedColumn = movedLeft ? column : this.getNextColumn(column);
        this.columns.move(column, prevColumn);
        this.columnsSetX(firstMovedColumn);
        this.desktop.onLayoutChanged();
        this.desktop.autoAdjustScroll();
    }
    moveColumnLeft(column) {
        this.columns.moveBack(column);
        this.columnsSetX(column);
        this.desktop.onLayoutChanged();
        this.desktop.autoAdjustScroll();
    }
    moveColumnRight(column) {
        const nextColumn = this.columns.getNext(column);
        if (nextColumn === null) {
            return;
        }
        this.moveColumnLeft(nextColumn);
    }
    getWidth() {
        return this.width;
    }
    isUserResizing() {
        return this.userResize;
    }
    getPrevColumn(column) {
        return this.columns.getPrev(column);
    }
    getNextColumn(column) {
        return this.columns.getNext(column);
    }
    getFirstColumn() {
        return this.columns.getFirst();
    }
    getLastColumn() {
        return this.columns.getLast();
    }
    getColumnAtIndex(i) {
        return this.columns.getItemAtIndex(i);
    }
    getLastFocusedColumn() {
        if (this.lastFocusedColumn === null || this.lastFocusedColumn.grid !== this) {
            return null;
        }
        return this.lastFocusedColumn;
    }
    getLastFocusedWindow() {
        const lastFocusedColumn = this.getLastFocusedColumn();
        if (lastFocusedColumn === null) {
            return null;
        }
        return lastFocusedColumn.getFocusTaker();
    }
    columnsSetX(firstMovedColumn) {
        const lastUnmovedColumn = firstMovedColumn === null ? this.columns.getLast() : this.columns.getPrev(firstMovedColumn);
        let x = lastUnmovedColumn === null ? 0 : lastUnmovedColumn.getRight() + this.config.gapsInnerHorizontal;
        if (firstMovedColumn !== null) {
            for (const column of this.columns.iteratorFrom(firstMovedColumn)) {
                column.gridX = x;
                x += column.getWidth() + this.config.gapsInnerHorizontal;
            }
        }
        this.width = x - this.config.gapsInnerHorizontal;
    }
    getLeftmostVisibleColumn(visibleRange, fullyVisible) {
        for (const column of this.columns.iterator()) {
            if (column.isVisible(visibleRange, fullyVisible)) {
                return column;
            }
        }
        return null;
    }
    getRightmostVisibleColumn(visibleRange, fullyVisible) {
        let last = null;
        for (const column of this.columns.iterator()) {
            if (column.isVisible(visibleRange, fullyVisible)) {
                last = column;
            }
            else if (last !== null) {
                break;
            }
        }
        return last;
    }
    *getVisibleColumns(visibleRange, fullyVisible) {
        for (const column of this.columns.iterator()) {
            if (column.isVisible(visibleRange, fullyVisible)) {
                yield column;
            }
        }
    }
    getVisibleColumnsWidth(visibleRange, fullyVisible) {
        let width = 0;
        let nVisible = 0;
        for (const column of this.getVisibleColumns(visibleRange, fullyVisible)) {
            width += column.getWidth();
            nVisible++;
        }
        if (nVisible > 0) {
            width += (nVisible - 1) * this.config.gapsInnerHorizontal;
        }
        return width;
    }
    arrange(x, visibleRange) {
        for (const column of this.columns.iterator()) {
            column.arrange(x, visibleRange, this.userResize);
            x += column.getWidth() + this.config.gapsInnerHorizontal;
        }
        const focusedWindow = this.getLastFocusedWindow();
        if (focusedWindow !== null) {
            focusedWindow.client.ensureTransientsVisible(this.desktop.clientArea);
        }
    }
    onColumnAdded(column, prevColumn) {
        if (prevColumn === null) {
            this.columns.insertStart(column);
        }
        else {
            this.columns.insertAfter(column, prevColumn);
        }
        this.columnsSetX(column);
        this.desktop.onLayoutChanged();
        this.desktop.autoAdjustScroll();
    }
    onColumnRemoved(column, passFocus) {
        const isLastColumn = this.columns.length() === 1;
        const nextColumn = this.getNextColumn(column);
        const columnToFocus = isLastColumn ? null : this.getPrevColumn(column) ?? nextColumn;
        if (column === this.lastFocusedColumn) {
            this.lastFocusedColumn = columnToFocus;
        }
        this.columns.remove(column);
        this.columnsSetX(nextColumn);
        this.desktop.onLayoutChanged();
        if (passFocus && columnToFocus !== null) {
            columnToFocus.focus();
        }
        else {
            this.desktop.autoAdjustScroll();
        }
    }
    onColumnWidthChanged(column) {
        const nextColumn = this.columns.getNext(column);
        this.columnsSetX(nextColumn);
        this.desktop.onLayoutChanged();
        if (!this.userResize) {
            this.desktop.autoAdjustScroll();
        }
    }
    onColumnFocused(column) {
        const lastFocusedColumn = this.getLastFocusedColumn();
        if (lastFocusedColumn !== null) {
            lastFocusedColumn.restoreToTiled();
        }
        this.lastFocusedColumn = column;
        this.desktop.scrollToColumn(column);
    }
    onScreenSizeChanged() {
        for (const column of this.columns.iterator()) {
            column.updateWidth();
            column.resizeWindows();
        }
    }
    onUserResizeStarted() {
        this.userResize = true;
    }
    onUserResizeFinished() {
        this.userResize = false;
        this.userResizeFinishedDelayer.run();
    }
    evacuateTail(targetGrid, startColumn) {
        for (const column of this.columns.iteratorFrom(startColumn)) {
            column.moveToGrid(targetGrid, targetGrid.getLastColumn());
        }
    }
    evacuate(targetGrid) {
        for (const column of this.columns.iterator()) {
            column.moveToGrid(targetGrid, targetGrid.getLastColumn());
        }
    }
    destroy() {
        this.userResizeFinishedDelayer.destroy();
    }
}
class Window {
    constructor(client, column) {
        this.client = client;
        this.height = client.kwinClient.frameGeometry.height;
        this.focusedState = {
            fullScreen: false,
            maximizedMode: 0 /* MaximizedMode.Unmaximized */,
        };
        this.skipArrange = false;
        this.column = column;
        column.onWindowAdded(this);
    }
    moveToColumn(targetColumn) {
        if (targetColumn === this.column) {
            return;
        }
        this.column.onWindowRemoved(this, false);
        this.column = targetColumn;
        targetColumn.onWindowAdded(this);
    }
    arrange(x, y, width, height) {
        if (this.skipArrange) {
            // window is maximized, fullscreen, or being manually resized, prevent fighting with the user
            return;
        }
        let maximized = false;
        if (this.column.grid.config.reMaximize && this.isFocused()) {
            // do this here rather than in `onFocused` to ensure it happens after placement
            // (otherwise placement may not happen at all)
            if (this.focusedState.maximizedMode !== 0 /* MaximizedMode.Unmaximized */) {
                this.client.setMaximize(this.focusedState.maximizedMode === 2 /* MaximizedMode.Horizontally */ || this.focusedState.maximizedMode === 3 /* MaximizedMode.Maximized */, this.focusedState.maximizedMode === 1 /* MaximizedMode.Vertically */ || this.focusedState.maximizedMode === 3 /* MaximizedMode.Maximized */);
                maximized = true;
            }
            if (this.focusedState.fullScreen) {
                this.client.setFullScreen(true);
                maximized = true;
            }
        }
        if (!maximized) {
            this.client.place(x, y, width, height);
        }
    }
    focus() {
        if (this.client.isShaded()) {
            // workaround for KWin deactivating clients when unshading immediately after activation
            this.client.setShade(false);
        }
        this.client.focus();
    }
    isFocused() {
        return this.client.isFocused();
    }
    onFocused() {
        this.column.onWindowFocused(this);
    }
    restoreToTiled() {
        if (this.isFocused()) {
            return;
        }
        this.client.setFullScreen(false);
        this.client.setMaximize(false, false);
        this.column.grid.desktop.onLayoutChanged();
    }
    onMaximizedChanged(maximizedMode) {
        const maximized = maximizedMode !== 0 /* MaximizedMode.Unmaximized */;
        this.skipArrange = maximized;
        if (this.column.grid.config.tiledKeepBelow) {
            this.client.kwinClient.keepBelow = !maximized;
        }
        if (this.column.grid.config.maximizedKeepAbove) {
            this.client.kwinClient.keepAbove = maximized;
        }
        if (this.isFocused()) {
            this.focusedState.maximizedMode = maximizedMode;
        }
        this.column.grid.desktop.onLayoutChanged();
    }
    onFullScreenChanged(fullScreen) {
        this.skipArrange = fullScreen;
        if (this.column.grid.config.tiledKeepBelow) {
            this.client.kwinClient.keepBelow = !fullScreen;
        }
        if (this.column.grid.config.maximizedKeepAbove) {
            this.client.kwinClient.keepAbove = fullScreen;
        }
        if (this.isFocused()) {
            this.focusedState.fullScreen = fullScreen;
        }
        this.column.grid.desktop.onLayoutChanged();
    }
    onUserResize(oldGeometry, resizeNeighborColumn) {
        const newGeometry = this.client.kwinClient.frameGeometry;
        const widthDelta = newGeometry.width - oldGeometry.width;
        const heightDelta = newGeometry.height - oldGeometry.height;
        if (widthDelta !== 0) {
            this.column.adjustWidth(widthDelta, true);
            let leftEdgeDelta = newGeometry.left - oldGeometry.left;
            const resizingLeftSide = leftEdgeDelta !== 0;
            if (resizeNeighborColumn && this.column.grid.config.resizeNeighborColumn) {
                const neighborColumn = resizingLeftSide ? this.column.grid.getPrevColumn(this.column) : this.column.grid.getNextColumn(this.column);
                if (neighborColumn !== null) {
                    const oldNeighborWidth = neighborColumn.getWidth();
                    neighborColumn.adjustWidth(-widthDelta, true);
                    if (resizingLeftSide) {
                        leftEdgeDelta -= neighborColumn.getWidth() - oldNeighborWidth;
                    }
                }
            }
            this.column.grid.desktop.adjustScroll(-leftEdgeDelta, true);
        }
        if (heightDelta !== 0) {
            this.column.adjustWindowHeight(this, heightDelta, newGeometry.y !== oldGeometry.y);
        }
    }
    onFrameGeometryChanged() {
        const newGeometry = this.client.kwinClient.frameGeometry;
        this.column.setWidth(newGeometry.width, true);
        this.column.grid.desktop.onLayoutChanged();
    }
    destroy(passFocus) {
        this.column.onWindowRemoved(this, passFocus);
    }
}
class ClientMatcher {
    constructor(regex) {
        this.regex = regex;
    }
    matches(kwinClient) {
        return this.regex.test(ClientMatcher.getClientString(kwinClient));
    }
    static getClientString(kwinClient) {
        return ClientMatcher.getRuleString(kwinClient.resourceClass, kwinClient.caption);
    }
    static getRuleString(ruleClass, ruleCaption) {
        return ruleClass + "\0" + ruleCaption;
    }
}
class WindowRuleEnforcer {
    constructor(windowRules) {
        const [floatRegex, tileRegex, followCaptionRegex] = WindowRuleEnforcer.createWindowRuleRegexes(windowRules);
        this.preferFloating = new ClientMatcher(floatRegex);
        this.preferTiling = new ClientMatcher(tileRegex);
        this.followCaption = followCaptionRegex;
    }
    shouldTile(kwinClient) {
        return Clients.canTileNow(kwinClient) && (this.preferTiling.matches(kwinClient) || (kwinClient.normalWindow &&
            !kwinClient.transient &&
            kwinClient.managed &&
            kwinClient.pid > -1 &&
            !this.preferFloating.matches(kwinClient)));
    }
    initClientSignalManager(world, kwinClient) {
        if (!this.followCaption.test(kwinClient.resourceClass)) {
            return null;
        }
        const enforcer = this;
        const manager = new SignalManager();
        manager.connect(kwinClient.captionChanged, () => {
            const shouldTile = enforcer.shouldTile(kwinClient);
            world.do((clientManager, desktopManager) => {
                const desktop = desktopManager.getDesktopForClient(kwinClient);
                if (shouldTile && desktop !== undefined) {
                    clientManager.tileKwinClient(kwinClient, desktop.grid);
                }
                else {
                    clientManager.floatKwinClient(kwinClient);
                }
            });
        });
        return manager;
    }
    static createWindowRuleRegexes(windowRules) {
        const floatRegexes = [];
        const tileRegexes = [];
        const followCaptionRegexes = [];
        for (const windowRule of windowRules) {
            const ruleClass = WindowRuleEnforcer.parseRegex(windowRule.class);
            const ruleCaption = WindowRuleEnforcer.parseRegex(windowRule.caption);
            const ruleString = ClientMatcher.getRuleString(WindowRuleEnforcer.wrapParens(ruleClass), WindowRuleEnforcer.wrapParens(ruleCaption));
            (windowRule.tile ? tileRegexes : floatRegexes).push(ruleString);
            if (ruleCaption !== ".*") {
                followCaptionRegexes.push(ruleClass);
            }
        }
        return [
            WindowRuleEnforcer.joinRegexes(floatRegexes),
            WindowRuleEnforcer.joinRegexes(tileRegexes),
            WindowRuleEnforcer.joinRegexes(followCaptionRegexes),
        ];
    }
    static parseRegex(rawRule) {
        if (rawRule === undefined || rawRule === "" || rawRule === ".*") {
            return ".*";
        }
        else {
            return rawRule;
        }
    }
    static joinRegexes(regexes) {
        if (regexes.length === 0) {
            return new RegExp("a^"); // match nothing
        }
        if (regexes.length === 1) {
            return new RegExp("^(" + regexes[0] + ")$");
        }
        const joinedRegexes = regexes.map(WindowRuleEnforcer.wrapParens).join("|");
        return new RegExp("^(" + joinedRegexes + ")$");
    }
    static wrapParens(str) {
        return "(" + str + ")";
    }
}
class Delayer {
    constructor(delay, f) {
        this.timer = initQmlTimer();
        this.timer.interval = delay;
        this.timer.triggered.connect(f);
    }
    run() {
        this.timer.restart();
    }
    destroy() {
        this.timer.destroy();
    }
}
function initQmlTimer() {
    return Qt.createQmlObject(`import QtQuick 6.0
        Timer {}`, qmlBase);
}
class Doer {
    constructor() {
        this.nCalls = 0;
    }
    do(f) {
        this.nCalls++;
        f();
        this.nCalls--;
    }
    isDoing() {
        return this.nCalls > 0;
    }
}
class LinkedList {
    constructor() {
        this.firstNode = null;
        this.lastNode = null;
        this.itemMap = new Map();
    }
    getNode(item) {
        const node = this.itemMap.get(item);
        if (node === undefined) {
            throw new Error("item not in list");
        }
        return node;
    }
    insertBefore(item, nextItem) {
        const nextNode = this.getNode(nextItem);
        this.insert(item, nextNode.prev, nextNode);
    }
    insertAfter(item, prevItem) {
        const prevNode = this.getNode(prevItem);
        this.insert(item, prevNode, prevNode.next);
    }
    insertStart(item) {
        this.insert(item, null, this.firstNode);
    }
    insertEnd(item) {
        this.insert(item, this.lastNode, null);
    }
    insert(item, prevNode, nextNode) {
        const node = new LinkedList.Node(item);
        this.itemMap.set(item, node);
        this.insertNode(node, prevNode, nextNode);
    }
    insertNode(node, prevNode, nextNode) {
        node.prev = prevNode;
        node.next = nextNode;
        if (nextNode !== null) {
            console.assert(nextNode.prev === prevNode);
            nextNode.prev = node;
        }
        if (prevNode !== null) {
            console.assert(prevNode.next === nextNode);
            prevNode.next = node;
        }
        if (this.firstNode === nextNode) {
            this.firstNode = node;
        }
        if (this.lastNode === prevNode) {
            this.lastNode = node;
        }
    }
    getPrev(item) {
        const prevNode = this.getNode(item).prev;
        return prevNode === null ? null : prevNode.item;
    }
    getNext(item) {
        const nextNode = this.getNode(item).next;
        return nextNode === null ? null : nextNode.item;
    }
    getFirst() {
        if (this.firstNode === null) {
            return null;
        }
        return this.firstNode.item;
    }
    getLast() {
        if (this.lastNode === null) {
            return null;
        }
        return this.lastNode.item;
    }
    getItemAtIndex(index) {
        let node = this.firstNode;
        if (node === null) {
            return null;
        }
        for (let i = 0; i < index; i++) {
            node = node.next;
            if (node === null) {
                return null;
            }
        }
        return node.item;
    }
    remove(item) {
        const node = this.getNode(item);
        this.itemMap.delete(item);
        this.removeNode(node);
    }
    removeNode(node) {
        const prevNode = node.prev;
        const nextNode = node.next;
        if (prevNode !== null) {
            prevNode.next = nextNode;
        }
        if (nextNode !== null) {
            nextNode.prev = prevNode;
        }
        if (this.firstNode === node) {
            this.firstNode = nextNode;
        }
        if (this.lastNode === node) {
            this.lastNode = prevNode;
        }
    }
    contains(item) {
        return this.itemMap.has(item);
    }
    swap(node0, node1) {
        console.assert(node0.next === node1 && node1.prev === node0);
        const prevNode = node0.prev;
        const nextNode = node1.next;
        if (prevNode !== null) {
            prevNode.next = node1;
        }
        node1.next = node0;
        node0.next = nextNode;
        if (nextNode !== null) {
            nextNode.prev = node0;
        }
        node0.prev = node1;
        node1.prev = prevNode;
        if (this.firstNode === node0) {
            this.firstNode = node1;
        }
        if (this.lastNode === node1) {
            this.lastNode = node0;
        }
    }
    move(item, prevItem) {
        const node = this.getNode(item);
        this.removeNode(node);
        if (prevItem === null) {
            this.insertNode(node, null, this.firstNode);
        }
        else {
            const prevNode = this.getNode(prevItem);
            this.insertNode(node, prevNode, prevNode.next);
        }
    }
    moveBack(item) {
        const node = this.getNode(item);
        if (node.prev !== null) {
            console.assert(node !== this.firstNode);
            this.swap(node.prev, node);
        }
    }
    moveForward(item) {
        const node = this.getNode(item);
        if (node.next !== null) {
            console.assert(node !== this.lastNode);
            this.swap(node, node.next);
        }
    }
    length() {
        return this.itemMap.size;
    }
    *iterator() {
        for (let node = this.firstNode; node !== null; node = node.next) {
            yield node.item;
        }
    }
    *iteratorFrom(startItem) {
        for (let node = this.getNode(startItem); node !== null; node = node.next) {
            yield node.item;
        }
    }
}
(function (LinkedList) {
    // TODO (optimization): reuse nodes
    class Node {
        constructor(item) {
            this.item = item;
            this.prev = null;
            this.next = null;
        }
    }
    LinkedList.Node = Node;
})(LinkedList || (LinkedList = {}));
class ShortcutAction {
    constructor(keyBinding, f) {
        this.shortcutHandler = ShortcutAction.initShortcutHandler(keyBinding);
        this.shortcutHandler.activated.connect(f);
    }
    destroy() {
        this.shortcutHandler.destroy();
    }
    static initShortcutHandler(keyBinding) {
        return Qt.createQmlObject(`import QtQuick 6.0
import org.kde.kwin 3.0
ShortcutHandler {
    name: "karousel-${keyBinding.name}";
    text: "Karousel: ${keyBinding.description}";
    sequence: "${keyBinding.defaultKeySequence}";
}`, qmlBase);
    }
}
class SignalManager {
    constructor() {
        this.connections = [];
    }
    connect(signal, handler) {
        signal.connect(handler);
        this.connections.push({ signal: signal, handler: handler });
    }
    destroy() {
        for (const connection of this.connections) {
            connection.signal.disconnect(connection.handler);
        }
        this.connections = [];
    }
}
function log(...args) {
    console.log("Karousel:", ...args);
}
function clamp(value, min, max) {
    if (value < min) {
        return min;
    }
    if (value > max) {
        return max;
    }
    return value;
}
function union(array0, array1) {
    const set = new Set([...array0, ...array1]);
    return [...set];
}
class ClientManager {
    constructor(config, world, desktopManager, pinManager) {
        this.world = world;
        this.config = { keepAbove: config.floatingKeepAbove };
        this.desktopManager = desktopManager;
        this.pinManager = pinManager;
        this.clientMap = new Map();
        this.lastFocusedClient = null;
        let parsedWindowRules = [];
        try {
            parsedWindowRules = JSON.parse(config.windowRules);
        }
        catch (error) {
            notificationInvalidWindowRules.sendEvent();
            log("failed to parse windowRules:", error);
        }
        this.windowRuleEnforcer = new WindowRuleEnforcer(parsedWindowRules);
    }
    addClient(kwinClient) {
        console.assert(!this.hasClient(kwinClient));
        const desktop = this.desktopManager.getDesktopForClient(kwinClient);
        let constructState;
        if (kwinClient.dock) {
            constructState = () => new ClientState.Docked(this.world, kwinClient);
        }
        else if (this.windowRuleEnforcer.shouldTile(kwinClient) && desktop !== undefined) {
            constructState = (client) => new ClientState.Tiled(this.world, client, desktop.grid);
        }
        else {
            constructState = (client) => new ClientState.Floating(this.world, client, this.config, false);
        }
        const client = new ClientWrapper(kwinClient, constructState, this.findTransientFor(kwinClient), this.windowRuleEnforcer.initClientSignalManager(this.world, kwinClient));
        this.clientMap.set(kwinClient, client);
    }
    removeClient(kwinClient, passFocus) {
        console.assert(this.hasClient(kwinClient));
        const client = this.clientMap.get(kwinClient);
        if (client === undefined) {
            return;
        }
        client.destroy(passFocus && kwinClient === this.lastFocusedClient);
        this.clientMap.delete(kwinClient);
    }
    findTransientFor(kwinClient) {
        if (!kwinClient.transient) {
            return null;
        }
        const transientFor = this.clientMap.get(kwinClient.transientFor);
        if (transientFor === undefined) {
            return null;
        }
        return transientFor;
    }
    minimizeClient(kwinClient) {
        const client = this.clientMap.get(kwinClient);
        if (client === undefined) {
            return;
        }
        if (client.stateManager.getState() instanceof ClientState.Tiled) {
            client.stateManager.setState(() => new ClientState.TiledMinimized(this.world, client), kwinClient === this.lastFocusedClient);
        }
    }
    tileClient(client, grid) {
        if (client.stateManager.getState() instanceof ClientState.Tiled) {
            return;
        }
        client.stateManager.setState(() => new ClientState.Tiled(this.world, client, grid), false);
    }
    floatClient(client) {
        if (client.stateManager.getState() instanceof ClientState.Floating) {
            return;
        }
        client.stateManager.setState(() => new ClientState.Floating(this.world, client, this.config, true), false);
    }
    tileKwinClient(kwinClient, grid) {
        const client = this.clientMap.get(kwinClient);
        if (client === undefined) {
            return;
        }
        this.tileClient(client, grid);
    }
    floatKwinClient(kwinClient) {
        const client = this.clientMap.get(kwinClient);
        if (client === undefined) {
            return;
        }
        this.floatClient(client);
    }
    pinClient(kwinClient) {
        const client = this.clientMap.get(kwinClient);
        if (client === undefined) {
            return;
        }
        if (client.getMaximizedMode() !== 0 /* MaximizedMode.Unmaximized */) {
            // the client is not really kwin-tiled, just maximized
            kwinClient.tile = null;
            return;
        }
        client.stateManager.setState(() => new ClientState.Pinned(this.world, this.pinManager, this.desktopManager, kwinClient, this.config), false);
        this.pinManager.addClient(kwinClient);
        for (const desktop of this.desktopManager.getDesktopsForClient(kwinClient)) {
            desktop.onPinsChanged();
        }
    }
    unpinClient(kwinClient) {
        const client = this.clientMap.get(kwinClient);
        if (client === undefined) {
            return;
        }
        console.assert(client.stateManager.getState() instanceof ClientState.Pinned);
        client.stateManager.setState(() => new ClientState.Floating(this.world, client, this.config, false), false);
        this.pinManager.removeClient(kwinClient);
        for (const desktop of this.desktopManager.getDesktopsForClient(kwinClient)) {
            desktop.onPinsChanged();
        }
    }
    toggleFloatingClient(kwinClient) {
        const client = this.clientMap.get(kwinClient);
        if (client === undefined) {
            return;
        }
        const clientState = client.stateManager.getState();
        if ((clientState instanceof ClientState.Floating || clientState instanceof ClientState.Pinned) && Clients.canTileEver(kwinClient)) {
            Clients.makeTileable(kwinClient);
            const desktop = this.desktopManager.getDesktopForClient(kwinClient);
            if (desktop === undefined) {
                return;
            }
            client.stateManager.setState(() => new ClientState.Tiled(this.world, client, desktop.grid), false);
        }
        else if (clientState instanceof ClientState.Tiled) {
            client.stateManager.setState(() => new ClientState.Floating(this.world, client, this.config, true), false);
        }
    }
    hasClient(kwinClient) {
        return this.clientMap.has(kwinClient);
    }
    onClientFocused(kwinClient) {
        this.lastFocusedClient = kwinClient;
        const window = this.findTiledWindow(kwinClient, true);
        if (window !== null) {
            window.onFocused();
        }
    }
    findTiledWindow(kwinClient, followTransient) {
        const client = this.clientMap.get(kwinClient);
        if (client === undefined) {
            return null;
        }
        return this.findTiledWindowOfClient(client, followTransient);
    }
    findTiledWindowOfClient(client, followTransient) {
        const clientState = client.stateManager.getState();
        if (clientState instanceof ClientState.Tiled) {
            return clientState.window;
        }
        else if (followTransient && client.transientFor !== null) {
            return this.findTiledWindowOfClient(client.transientFor, true);
        }
        else {
            return null;
        }
    }
    removeAllClients() {
        for (const kwinClient of Array.from(this.clientMap.keys())) {
            this.removeClient(kwinClient, false);
        }
    }
    destroy() {
        this.removeAllClients();
    }
}
class ClientWrapper {
    constructor(kwinClient, constructInitialState, transientFor, rulesSignalManager) {
        this.kwinClient = kwinClient;
        this.transientFor = transientFor;
        this.transients = [];
        if (transientFor !== null) {
            transientFor.addTransient(this);
        }
        this.signalManager = ClientWrapper.initSignalManager(this);
        this.rulesSignalManager = rulesSignalManager;
        this.preferredWidth = kwinClient.frameGeometry.width;
        this.manipulatingGeometry = new Doer();
        this.lastPlacement = null;
        this.stateManager = new ClientState.Manager(constructInitialState(this));
    }
    place(x, y, width, height) {
        this.manipulatingGeometry.do(() => {
            if (this.kwinClient.resize) {
                // window is being manually resized, prevent fighting with the user
                return;
            }
            this.lastPlacement = Qt.rect(x, y, width, height);
            this.kwinClient.frameGeometry = this.lastPlacement;
            if (this.kwinClient.frameGeometry !== this.lastPlacement) {
                // frameGeometry assignment failed. This sometimes happens on Wayland
                // when a window is off-screen, effectively making it stuck there.
                this.kwinClient.frameGeometry.x = x; // This makes it unstuck.
                this.kwinClient.frameGeometry = this.lastPlacement;
            }
        });
    }
    moveTransient(dx, dy, kwinDesktops) {
        if (this.stateManager.getState() instanceof ClientState.Floating) {
            if (Clients.isOnOneOfVirtualDesktops(this.kwinClient, kwinDesktops)) {
                const frame = this.kwinClient.frameGeometry;
                this.kwinClient.frameGeometry = Qt.rect(frame.x + dx, frame.y + dy, frame.width, frame.height);
            }
            for (const transient of this.transients) {
                transient.moveTransient(dx, dy, kwinDesktops);
            }
        }
    }
    moveTransients(dx, dy) {
        for (const transient of this.transients) {
            transient.moveTransient(dx, dy, this.kwinClient.desktops);
        }
    }
    focus() {
        Workspace.activeWindow = this.kwinClient;
    }
    isFocused() {
        return Workspace.activeWindow === this.kwinClient;
    }
    setMaximize(horizontally, vertically) {
        if (!this.kwinClient.maximizable) {
            return;
        }
        if (this.maximizedMode === undefined) {
            if (horizontally && vertically) {
                this.maximizedMode = 3 /* MaximizedMode.Maximized */;
            }
            else if (horizontally) {
                this.maximizedMode = 2 /* MaximizedMode.Horizontally */;
            }
            else if (vertically) {
                this.maximizedMode = 1 /* MaximizedMode.Vertically */;
            }
            else {
                this.maximizedMode = 0 /* MaximizedMode.Unmaximized */;
            }
        }
        this.manipulatingGeometry.do(() => {
            this.kwinClient.setMaximize(vertically, horizontally);
        });
    }
    setFullScreen(fullScreen) {
        if (!this.kwinClient.fullScreenable) {
            return;
        }
        this.manipulatingGeometry.do(() => {
            this.kwinClient.fullScreen = fullScreen;
        });
    }
    setShade(shade) {
        this.manipulatingGeometry.do(() => {
            this.kwinClient.shade = shade;
        });
    }
    isShaded() {
        return this.kwinClient.shade;
    }
    getMaximizedMode() {
        return this.maximizedMode;
    }
    isManipulatingGeometry(newGeometry) {
        if (newGeometry !== null && newGeometry === this.lastPlacement) {
            return true;
        }
        return this.manipulatingGeometry.isDoing();
    }
    addTransient(transient) {
        this.transients.push(transient);
    }
    removeTransient(transient) {
        const i = this.transients.indexOf(transient);
        this.transients.splice(i, 1);
    }
    ensureTransientsVisible(screenSize) {
        for (const transient of this.transients) {
            if (transient.stateManager.getState() instanceof ClientState.Floating) {
                transient.ensureVisible(screenSize);
                transient.ensureTransientsVisible(screenSize);
            }
        }
    }
    ensureVisible(screenSize) {
        if (!Clients.isOnVirtualDesktop(this.kwinClient, Workspace.currentDesktop)) {
            return;
        }
        const frame = this.kwinClient.frameGeometry;
        if (frame.left < screenSize.left) {
            frame.x = screenSize.left;
        }
        else if (frame.right > screenSize.right) {
            frame.x = screenSize.right - frame.width;
        }
    }
    destroy(passFocus) {
        this.stateManager.destroy(passFocus);
        this.signalManager.destroy();
        if (this.rulesSignalManager !== null) {
            this.rulesSignalManager.destroy();
        }
        if (this.transientFor !== null) {
            this.transientFor.removeTransient(this);
        }
        for (const transient of this.transients) {
            transient.transientFor = null;
        }
    }
    static initSignalManager(client) {
        const manager = new SignalManager();
        manager.connect(client.kwinClient.maximizedAboutToChange, (maximizedMode) => {
            if (maximizedMode !== 0 /* MaximizedMode.Unmaximized */ && client.kwinClient.tile !== null) {
                client.kwinClient.tile = null;
            }
            client.maximizedMode = maximizedMode;
        });
        return manager;
    }
}
var Clients;
(function (Clients) {
    function canTileEver(kwinClient) {
        return kwinClient.moveable && kwinClient.resizeable && !kwinClient.popupWindow;
    }
    Clients.canTileEver = canTileEver;
    function canTileNow(kwinClient) {
        return canTileEver(kwinClient) && !kwinClient.minimized && kwinClient.desktops.length === 1 && kwinClient.activities.length === 1;
    }
    Clients.canTileNow = canTileNow;
    function makeTileable(kwinClient) {
        if (kwinClient.minimized) {
            kwinClient.minimized = false;
        }
        if (kwinClient.desktops.length !== 1) {
            kwinClient.desktops = [Workspace.currentDesktop];
        }
        if (kwinClient.activities.length !== 1) {
            kwinClient.activities = [Workspace.currentActivity];
        }
    }
    Clients.makeTileable = makeTileable;
    function getKwinDesktopApprox(kwinClient) {
        switch (kwinClient.desktops.length) {
            case 0:
                return Workspace.currentDesktop;
            case 1:
                return kwinClient.desktops[0];
            default:
                if (kwinClient.desktops.includes(Workspace.currentDesktop)) {
                    return Workspace.currentDesktop;
                }
                else {
                    return kwinClient.desktops[0];
                }
        }
    }
    Clients.getKwinDesktopApprox = getKwinDesktopApprox;
    function isFullScreenGeometry(kwinClient) {
        const fullScreenArea = Workspace.clientArea(4 /* ClientAreaOption.FullScreenArea */, kwinClient.output, getKwinDesktopApprox(kwinClient));
        return kwinClient.frameGeometry === fullScreenArea;
    }
    Clients.isFullScreenGeometry = isFullScreenGeometry;
    function isOnVirtualDesktop(kwinClient, kwinDesktop) {
        return kwinClient.desktops.length === 0 || kwinClient.desktops.includes(kwinDesktop);
    }
    Clients.isOnVirtualDesktop = isOnVirtualDesktop;
    function isOnOneOfVirtualDesktops(kwinClient, kwinDesktops) {
        return kwinClient.desktops.length === 0 || kwinClient.desktops.some(d => kwinDesktops.includes(d));
    }
    Clients.isOnOneOfVirtualDesktops = isOnOneOfVirtualDesktops;
})(Clients || (Clients = {}));
class DesktopManager {
    constructor(pinManager, config, layoutConfig, currentActivity, currentDesktop) {
        this.pinManager = pinManager;
        this.config = config;
        this.layoutConfig = layoutConfig;
        this.desktops = new Map();
        this.selectedScreen = Workspace.activeScreen;
        this.kwinActivities = new Set(Workspace.activities);
        this.kwinDesktops = new Set(Workspace.desktops);
        this.addDesktop(currentActivity, currentDesktop);
    }
    getDesktop(activity, kwinDesktop) {
        const desktopKey = DesktopManager.getDesktopKey(activity, kwinDesktop);
        const desktop = this.desktops.get(desktopKey);
        if (desktop !== undefined) {
            return desktop;
        }
        else {
            return this.addDesktop(activity, kwinDesktop);
        }
    }
    getCurrentDesktop() {
        return this.getDesktop(Workspace.currentActivity, Workspace.currentDesktop);
    }
    getDesktopInCurrentActivity(kwinDesktop) {
        return this.getDesktop(Workspace.currentActivity, kwinDesktop);
    }
    getDesktopForClient(kwinClient) {
        if (kwinClient.activities.length !== 1 || kwinClient.desktops.length !== 1) {
            return undefined;
        }
        return this.getDesktop(kwinClient.activities[0], kwinClient.desktops[0]);
    }
    addDesktop(activity, kwinDesktop) {
        const desktopKey = DesktopManager.getDesktopKey(activity, kwinDesktop);
        const desktop = new Desktop(kwinDesktop, this.pinManager, this.config, () => this.selectedScreen, this.layoutConfig);
        this.desktops.set(desktopKey, desktop);
        return desktop;
    }
    static getDesktopKey(activity, kwinDesktop) {
        return activity + "|" + kwinDesktop.id;
    }
    updateActivities() {
        const newActivities = new Set(Workspace.activities);
        for (const activity of this.kwinActivities) {
            if (!newActivities.has(activity)) {
                this.removeActivity(activity);
            }
        }
        this.kwinActivities = newActivities;
    }
    updateDesktops() {
        const newDesktops = new Set(Workspace.desktops);
        for (const desktop of this.kwinDesktops) {
            if (!newDesktops.has(desktop)) {
                this.removeKwinDesktop(desktop);
            }
        }
        this.kwinDesktops = newDesktops;
    }
    selectScreen(screen) {
        this.selectedScreen = screen;
    }
    removeActivity(activity) {
        for (const kwinDesktop of this.kwinDesktops) {
            this.destroyDesktop(activity, kwinDesktop);
        }
    }
    removeKwinDesktop(kwinDesktop) {
        for (const activity of this.kwinActivities) {
            this.destroyDesktop(activity, kwinDesktop);
        }
    }
    destroyDesktop(activity, kwinDesktop) {
        const desktopKey = DesktopManager.getDesktopKey(activity, kwinDesktop);
        const desktop = this.desktops.get(desktopKey);
        if (desktop !== undefined) {
            desktop.destroy();
            this.desktops.delete(desktopKey);
        }
    }
    destroy() {
        for (const desktop of this.desktops.values()) {
            desktop.destroy();
        }
    }
    *getAllDesktops() {
        for (const desktop of this.desktops.values()) {
            yield desktop;
        }
    }
    getDesktopsForClient(kwinClient) {
        const desktops = this.getDesktops(kwinClient.activities, kwinClient.desktops); // workaround for QTBUG-109880
        return desktops;
    }
    // empty array means all
    *getDesktops(activities, kwinDesktops) {
        const matchedActivities = activities.length > 0 ? activities : this.kwinActivities.keys();
        const matchedDesktops = kwinDesktops.length > 0 ? kwinDesktops : this.kwinDesktops.keys();
        for (const matchedActivity of matchedActivities) {
            for (const matchedDesktop of matchedDesktops) {
                const desktopKey = DesktopManager.getDesktopKey(matchedActivity, matchedDesktop);
                const desktop = this.desktops.get(desktopKey);
                if (desktop !== undefined) {
                    yield desktop;
                }
            }
        }
    }
}
class PinManager {
    constructor() {
        this.pinnedClients = new Set();
    }
    addClient(kwinClient) {
        this.pinnedClients.add(kwinClient);
    }
    removeClient(kwinClient) {
        this.pinnedClients.delete(kwinClient);
    }
    getAvailableSpace(kwinDesktop, screen) {
        const baseLot = new PinManager.Lot(screen.top, screen.bottom, screen.left, screen.right);
        let lots = [baseLot];
        for (const client of this.pinnedClients) {
            if (!Clients.isOnVirtualDesktop(client, kwinDesktop)) {
                continue;
            }
            const newLots = [];
            for (const lot of lots) {
                lot.split(newLots, client.frameGeometry);
            }
            lots = newLots;
        }
        let largestLot = baseLot;
        let largestArea = 0;
        for (const lot of lots) {
            const area = lot.area();
            if (area > largestArea) {
                largestArea = area;
                largestLot = lot;
            }
        }
        return largestLot;
    }
}
(function (PinManager) {
    class Lot {
        constructor(top, bottom, left, right) {
            this.top = top;
            this.bottom = bottom;
            this.left = left;
            this.right = right;
        }
        split(destLots, obstacle) {
            if (!this.contains(obstacle)) {
                // don't split
                destLots.push(this);
                return;
            }
            if (obstacle.top - this.top >= Lot.minHeight) {
                destLots.push(new Lot(this.top, obstacle.top, this.left, this.right));
            }
            if (this.bottom - obstacle.bottom >= Lot.minHeight) {
                destLots.push(new Lot(obstacle.bottom, this.bottom, this.left, this.right));
            }
            if (obstacle.left - this.left >= Lot.minWidth) {
                destLots.push(new Lot(this.top, this.bottom, this.left, obstacle.left));
            }
            if (this.right - obstacle.right >= Lot.minWidth) {
                destLots.push(new Lot(this.top, this.bottom, obstacle.right, this.right));
            }
        }
        contains(obstacle) {
            return obstacle.right > this.left && obstacle.left < this.right &&
                obstacle.bottom > this.top && obstacle.top < this.bottom;
        }
        area() {
            return (this.bottom - this.top) * (this.right - this.left);
        }
    }
    Lot.minWidth = 200;
    Lot.minHeight = 200;
    PinManager.Lot = Lot;
})(PinManager || (PinManager = {}));
class World {
    constructor(config) {
        this.untileOnDrag = config.untileOnDrag;
        this.workspaceSignalManager = initWorkspaceSignalHandlers(this);
        this.shortcutActions = registerKeyBindings(this, {
            manualScrollStep: config.manualScrollStep,
            manualResizeStep: config.manualResizeStep,
            columnResizer: config.scrollingCentered ? new RawResizer() : new ContextualResizer(),
        });
        this.screenResizedDelayer = new Delayer(1000, () => {
            // this delay ensures that docks are taken into account by `Workspace.clientArea`
            for (const desktop of this.desktopManager.getAllDesktops()) {
                desktop.onLayoutChanged();
            }
            this.update();
        });
        this.pinManager = new PinManager();
        const layoutConfig = {
            gapsInnerHorizontal: config.gapsInnerHorizontal,
            gapsInnerVertical: config.gapsInnerVertical,
            offScreenOpacity: config.offScreenOpacity / 100.0,
            stackColumnsByDefault: config.stackColumnsByDefault,
            resizeNeighborColumn: config.resizeNeighborColumn,
            reMaximize: config.reMaximize,
            skipSwitcher: config.skipSwitcher,
            tiledKeepBelow: config.tiledKeepBelow,
            maximizedKeepAbove: config.floatingKeepAbove,
        };
        this.desktopManager = new DesktopManager(this.pinManager, {
            marginTop: config.gapsOuterTop,
            marginBottom: config.gapsOuterBottom,
            marginLeft: config.gapsOuterLeft,
            marginRight: config.gapsOuterRight,
            scroller: World.createScroller(config),
            clamper: config.scrollingLazy ? new EdgeClamper() : new CenterClamper(),
        }, layoutConfig, Workspace.currentActivity, Workspace.currentDesktop);
        this.clientManager = new ClientManager(config, this, this.desktopManager, this.pinManager);
        this.addExistingClients();
        this.update();
    }
    static createScroller(config) {
        if (config.scrollingLazy) {
            return new LazyScroller();
        }
        else if (config.scrollingCentered) {
            return new CenteredScroller();
        }
        else if (config.scrollingGrouped) {
            return new GroupedScroller();
        }
        else {
            log("No scrolling mode selected, using default");
            return new LazyScroller();
        }
    }
    addExistingClients() {
        const kwinClients = Workspace.windows;
        for (let i = 0; i < kwinClients.length; i++) {
            const kwinClient = kwinClients[i];
            this.clientManager.addClient(kwinClient);
        }
    }
    update() {
        this.desktopManager.getCurrentDesktop().arrange();
    }
    do(f) {
        f(this.clientManager, this.desktopManager);
        this.update();
    }
    doIfTiled(kwinClient, followTransient, f) {
        const window = this.clientManager.findTiledWindow(kwinClient, followTransient);
        if (window === null) {
            return;
        }
        const column = window.column;
        const grid = column.grid;
        f(this.clientManager, this.desktopManager, window, column, grid);
        this.update();
    }
    doIfTiledFocused(followTransient, f) {
        this.doIfTiled(Workspace.activeWindow, followTransient, f);
    }
    destroy() {
        this.workspaceSignalManager.destroy();
        for (const shortcutAction of this.shortcutActions) {
            shortcutAction.destroy();
        }
        this.clientManager.destroy();
        this.desktopManager.destroy();
    }
    onScreenResized() {
        this.screenResizedDelayer.run();
    }
}
var ClientState;
(function (ClientState) {
    class Docked {
        constructor(world, kwinClient) {
            this.world = world;
            this.signalManager = Docked.initSignalManager(world, kwinClient);
            world.onScreenResized();
        }
        destroy(passFocus) {
            this.signalManager.destroy();
            this.world.onScreenResized();
        }
        static initSignalManager(world, kwinClient) {
            const manager = new SignalManager();
            manager.connect(kwinClient.frameGeometryChanged, () => {
                world.onScreenResized();
            });
            return manager;
        }
    }
    ClientState.Docked = Docked;
})(ClientState || (ClientState = {}));
var ClientState;
(function (ClientState) {
    class Floating {
        constructor(world, client, config, limitHeight) {
            this.client = client;
            this.config = config;
            if (config.keepAbove) {
                client.kwinClient.keepAbove = true;
            }
            if (limitHeight && client.kwinClient.tile === null) {
                Floating.limitHeight(client);
            }
            this.signalManager = Floating.initSignalManager(world, client.kwinClient);
        }
        destroy(passFocus) {
            this.signalManager.destroy();
            if (this.config.keepAbove) {
                this.client.kwinClient.keepAbove = false;
            }
        }
        static limitHeight(client) {
            const placementArea = Workspace.clientArea(0 /* ClientAreaOption.PlacementArea */, client.kwinClient.output, Clients.getKwinDesktopApprox(client.kwinClient));
            const clientRect = client.kwinClient.frameGeometry;
            const width = client.preferredWidth;
            client.place(clientRect.x, clientRect.y, width, Math.min(clientRect.height, Math.round(placementArea.height / 2)));
        }
        static initSignalManager(world, kwinClient) {
            const manager = new SignalManager();
            manager.connect(kwinClient.tileChanged, () => {
                // on X11, this fires after `frameGeometryChanged`
                if (kwinClient.tile !== null) {
                    world.do((clientManager, desktopManager) => {
                        clientManager.pinClient(kwinClient);
                    });
                }
            });
            manager.connect(kwinClient.frameGeometryChanged, () => {
                // on Wayland, this fires after `tileChanged`
                if (kwinClient.tile !== null) {
                    world.do((clientManager, desktopManager) => {
                        clientManager.pinClient(kwinClient);
                    });
                }
            });
            return manager;
        }
    }
    ClientState.Floating = Floating;
})(ClientState || (ClientState = {}));
var ClientState;
(function (ClientState) {
    class Manager {
        constructor(initialState) {
            this.state = initialState;
        }
        setState(constructNewState, passFocus) {
            this.state.destroy(passFocus);
            this.state = constructNewState();
        }
        getState() {
            return this.state;
        }
        destroy(passFocus) {
            this.state.destroy(passFocus);
        }
    }
    ClientState.Manager = Manager;
})(ClientState || (ClientState = {}));
var ClientState;
(function (ClientState) {
    class Pinned {
        constructor(world, pinManager, desktopManager, kwinClient, config) {
            this.kwinClient = kwinClient;
            this.pinManager = pinManager;
            this.desktopManager = desktopManager;
            this.config = config;
            if (config.keepAbove) {
                kwinClient.keepAbove = true;
            }
            this.signalManager = Pinned.initSignalManager(world, pinManager, kwinClient);
        }
        destroy(passFocus) {
            this.signalManager.destroy();
            if (this.config.keepAbove) {
                this.kwinClient.keepAbove = true;
            }
            this.pinManager.removeClient(this.kwinClient);
            for (const desktop of this.desktopManager.getDesktopsForClient(this.kwinClient)) {
                desktop.onPinsChanged();
            }
        }
        static initSignalManager(world, pinManager, kwinClient) {
            const manager = new SignalManager();
            let oldActivities = kwinClient.activities;
            let oldDesktops = kwinClient.desktops;
            manager.connect(kwinClient.tileChanged, () => {
                if (kwinClient.tile === null) {
                    world.do((clientManager, desktopManager) => {
                        clientManager.unpinClient(kwinClient);
                    });
                }
            });
            manager.connect(kwinClient.frameGeometryChanged, () => {
                if (kwinClient.tile === null) {
                    world.do((clientManager, desktopManager) => {
                        clientManager.unpinClient(kwinClient);
                    });
                    return;
                }
                world.do((clientManager, desktopManager) => {
                    for (const desktop of desktopManager.getDesktopsForClient(kwinClient)) {
                        desktop.onPinsChanged();
                    }
                });
            });
            manager.connect(kwinClient.desktopsChanged, () => {
                const changedDesktops = oldDesktops.length === 0 || kwinClient.desktops.length === 0 ?
                    [] :
                    union(oldDesktops, kwinClient.desktops);
                world.do((clientManager, desktopManager) => {
                    for (const desktop of desktopManager.getDesktops(kwinClient.activities, changedDesktops)) {
                        desktop.onPinsChanged();
                    }
                });
                oldDesktops = kwinClient.desktops;
            });
            manager.connect(kwinClient.activitiesChanged, () => {
                const changedActivities = oldActivities.length === 0 || kwinClient.activities.length === 0 ?
                    [] :
                    union(oldActivities, kwinClient.activities);
                world.do((clientManager, desktopManager) => {
                    for (const desktop of desktopManager.getDesktops(changedActivities, kwinClient.desktops)) {
                        desktop.onPinsChanged();
                    }
                });
                oldActivities = kwinClient.activities;
            });
            return manager;
        }
    }
    ClientState.Pinned = Pinned;
})(ClientState || (ClientState = {}));
var ClientState;
(function (ClientState) {
    class Tiled {
        constructor(world, client, grid) {
            this.defaultState = { skipSwitcher: client.kwinClient.skipSwitcher };
            Tiled.prepareClientForTiling(client, grid.config);
            const column = new Column(grid, grid.getLastFocusedColumn() ?? grid.getLastColumn());
            const window = new Window(client, column);
            this.window = window;
            this.signalManager = Tiled.initSignalManager(world, window);
        }
        destroy(passFocus) {
            this.signalManager.destroy();
            const window = this.window;
            const grid = window.column.grid;
            const client = window.client;
            window.destroy(passFocus);
            Tiled.restoreClientAfterTiling(client, grid.config, this.defaultState, grid.desktop.clientArea);
        }
        static initSignalManager(world, window) {
            const client = window.client;
            const kwinClient = client.kwinClient;
            const manager = new SignalManager();
            manager.connect(kwinClient.desktopsChanged, () => {
                world.do((clientManager, desktopManager) => {
                    const desktop = desktopManager.getDesktopForClient(kwinClient);
                    if (desktop === undefined) {
                        // windows on multiple desktops are not supported
                        clientManager.floatKwinClient(kwinClient);
                        return;
                    }
                    Tiled.moveWindowToGrid(window, desktop.grid);
                });
            });
            manager.connect(kwinClient.activitiesChanged, () => {
                world.do((clientManager, desktopManager) => {
                    const desktop = desktopManager.getDesktopForClient(kwinClient);
                    if (desktop === undefined) {
                        // windows on multiple activities are not supported
                        clientManager.floatKwinClient(kwinClient);
                        return;
                    }
                    Tiled.moveWindowToGrid(window, desktop.grid);
                });
            });
            manager.connect(kwinClient.minimizedChanged, () => {
                console.assert(kwinClient.minimized);
                world.do((clientManager, desktopManager) => {
                    clientManager.minimizeClient(kwinClient);
                });
            });
            manager.connect(kwinClient.maximizedAboutToChange, (maximizedMode) => {
                world.do(() => {
                    window.onMaximizedChanged(maximizedMode);
                });
            });
            let resizing = false;
            let resizingBorder = false;
            manager.connect(kwinClient.interactiveMoveResizeStarted, () => {
                if (kwinClient.move) {
                    if (world.untileOnDrag) {
                        world.do((clientManager, desktopManager) => {
                            clientManager.floatKwinClient(kwinClient);
                        });
                    }
                    return;
                }
                if (kwinClient.resize) {
                    resizing = true;
                    resizingBorder = Workspace.cursorPos.x > kwinClient.clientGeometry.right ||
                        Workspace.cursorPos.x < kwinClient.clientGeometry.left;
                    window.column.grid.onUserResizeStarted();
                }
            });
            manager.connect(kwinClient.interactiveMoveResizeFinished, () => {
                if (resizing) {
                    resizing = false;
                    window.column.grid.onUserResizeFinished();
                }
            });
            manager.connect(kwinClient.frameGeometryChanged, (oldGeometry) => {
                // on Wayland, this fires after `tileChanged`
                if (kwinClient.tile !== null) {
                    world.do((clientManager, desktopManager) => {
                        clientManager.pinClient(kwinClient);
                    });
                    return;
                }
                const newGeometry = client.kwinClient.frameGeometry;
                const oldCenterX = oldGeometry.x + oldGeometry.width / 2;
                const oldCenterY = oldGeometry.y + oldGeometry.height / 2;
                const newCenterX = newGeometry.x + newGeometry.width / 2;
                const newCenterY = newGeometry.y + newGeometry.height / 2;
                const dx = Math.round(newCenterX - oldCenterX);
                const dy = Math.round(newCenterY - oldCenterY);
                if (dx !== 0 || dy !== 0) {
                    client.moveTransients(dx, dy);
                }
                if (kwinClient.resize) {
                    world.do(() => window.onUserResize(oldGeometry, resizingBorder));
                }
                else if (!window.column.grid.isUserResizing() &&
                    !client.isManipulatingGeometry(newGeometry) &&
                    client.getMaximizedMode() === 0 /* MaximizedMode.Unmaximized */ &&
                    !Clients.isFullScreenGeometry(kwinClient) // not using `kwinClient.fullScreen` because it may not be set yet at this point
                ) {
                    world.do(() => window.onFrameGeometryChanged());
                }
            });
            manager.connect(kwinClient.fullScreenChanged, () => {
                world.do(() => window.onFullScreenChanged(kwinClient.fullScreen));
            });
            manager.connect(kwinClient.tileChanged, () => {
                // on X11, this fires after `frameGeometryChanged`
                if (kwinClient.tile !== null) {
                    world.do((clientManager, desktopManager) => {
                        clientManager.pinClient(kwinClient);
                    });
                }
            });
            return manager;
        }
        static moveWindowToGrid(window, grid) {
            if (grid === window.column.grid) {
                // window already on the given grid
                return;
            }
            const newColumn = new Column(grid, grid.getLastFocusedColumn() ?? grid.getLastColumn());
            window.moveToColumn(newColumn);
        }
        static prepareClientForTiling(client, config) {
            if (config.skipSwitcher) {
                client.kwinClient.skipSwitcher = true;
            }
            if (config.tiledKeepBelow) {
                client.kwinClient.keepBelow = true;
            }
            client.setFullScreen(false);
            if (client.kwinClient.tile !== null) {
                client.setMaximize(false, true); // disable quick tile mode
            }
            client.setMaximize(false, false);
        }
        static restoreClientAfterTiling(client, config, defaultState, screenSize) {
            if (config.skipSwitcher) {
                client.kwinClient.skipSwitcher = defaultState.skipSwitcher;
            }
            if (config.tiledKeepBelow) {
                client.kwinClient.keepBelow = false;
            }
            if (config.offScreenOpacity < 1.0) {
                client.kwinClient.opacity = 1.0;
            }
            client.setShade(false);
            client.setFullScreen(false);
            if (client.kwinClient.tile === null) {
                client.setMaximize(false, false);
            }
            client.ensureVisible(screenSize);
        }
    }
    ClientState.Tiled = Tiled;
})(ClientState || (ClientState = {}));
var ClientState;
(function (ClientState) {
    class TiledMinimized {
        constructor(world, client) {
            this.signalManager = TiledMinimized.initSignalManager(world, client);
        }
        destroy(passFocus) {
            this.signalManager.destroy();
        }
        static initSignalManager(world, client) {
            const manager = new SignalManager();
            manager.connect(client.kwinClient.minimizedChanged, () => {
                console.assert(!client.kwinClient.minimized);
                world.do((clientManager, desktopManager) => {
                    const desktop = desktopManager.getDesktopForClient(client.kwinClient);
                    if (desktop !== undefined) {
                        clientManager.tileClient(client, desktop.grid);
                    }
                    else {
                        clientManager.floatClient(client);
                    }
                });
            });
            return manager;
        }
    }
    ClientState.TiledMinimized = TiledMinimized;
})(ClientState || (ClientState = {}));
function init() {
    return new World(loadConfig());
}
